import logging
from urllib.parse import urlparse

import aiohttp

from bovine.crypto.helper import content_digest_sha256
from bovine.crypto.http_signature import build_signature
from bovine.utils.date import get_gmt_now

from .consts import BOVINE_CLIENT_NAME
from .event_source import EventSource

logger = logging.getLogger(__name__)


async def signed_get(
    session: aiohttp.ClientSession,
    public_key_url: str,
    private_key: str,
    url: str,
    headers: dict = {},
) -> aiohttp.ClientResponse:
    logger.debug(f"Signed get with {public_key_url} on {url}")

    parsed_url = urlparse(url)
    host = parsed_url.netloc
    target = parsed_url.path

    accept = "application/activity+json"
    content_type = "application/activity+json"
    date_header = get_gmt_now()

    signature_header = (
        build_signature(host, "get", target)
        .with_field("date", date_header)
        .with_field("accept", accept)
        .build_signature(public_key_url, private_key)
    )

    headers["accept"] = accept
    headers["date"] = date_header
    headers["host"] = host
    headers["content-type"] = content_type
    headers["signature"] = signature_header
    headers["user-agent"] = BOVINE_CLIENT_NAME
    return await session.get(url, headers=headers)


def signed_event_source(
    session: aiohttp.ClientSession,
    public_key_url: str,
    private_key: str,
    url: str,
    headers: dict = {},
):
    logger.debug(f"Signed event source with {public_key_url} on {url}")

    parsed_url = urlparse(url)
    host = parsed_url.netloc
    target = parsed_url.path

    accept = "text/event-stream"
    date_header = get_gmt_now()

    signature_header = (
        build_signature(host, "get", target)
        .with_field("date", date_header)
        .with_field("accept", accept)
        .build_signature(public_key_url, private_key)
    )

    headers["accept"] = accept
    headers["date"] = date_header
    headers["host"] = host
    headers["signature"] = signature_header
    headers["user-agent"] = BOVINE_CLIENT_NAME
    return EventSource(session, url, headers=headers)


async def signed_post(
    session: aiohttp.ClientSession,
    public_key_url: str,
    private_key: str,
    url: str,
    body: str,
    headers: dict = {},
    content_type=None,
) -> aiohttp.ClientResponse:
    logger.debug(f"Signed post with {public_key_url} on {url}")

    parsed_url = urlparse(url)
    host = parsed_url.netloc
    target = parsed_url.path

    accept = "application/activity+json"
    # LABEL: ap-s2s-content-type
    if content_type is None:
        content_type = "application/activity+json"
    date_header = get_gmt_now()

    digest = content_digest_sha256(body)

    signature_header = (
        build_signature(host, "post", target)
        .with_field("date", date_header)
        .with_field("digest", digest)
        .with_field("content-type", content_type)
        .build_signature(public_key_url, private_key)
    )

    headers["accept"] = accept
    headers["digest"] = digest
    headers["date"] = date_header
    headers["host"] = host
    headers["content-type"] = content_type
    headers["signature"] = signature_header
    headers["user-agent"] = BOVINE_CLIENT_NAME

    return await session.post(url, data=body, headers=headers)



from . import parse_fediverse_handle


def test_parse_fediverse_handle():
    assert parse_fediverse_handle("account") == ("account", None)
    assert parse_fediverse_handle("account@domain") == ("account", "domain")

    assert parse_fediverse_handle("account@domain@@@") == ("account", "domain@@@")
    assert parse_fediverse_handle("@account@domain@@@") == ("account", "domain@@@")
    assert parse_fediverse_handle("@account") == ("account", None)



class CollectionIterator:
    def __init__(self, collection_helper, max_number: int):
        self.count = 0
        self.collection_helper = collection_helper
        self.max_number = max_number

    def __aiter__(self):
        return self

    async def __anext__(self):
        self.count += 1
        if self.count > self.max_number:
            raise StopAsyncIteration

        return await self.collection_helper.next_item()



import logging

logger = logging.getLogger(__name__)


class Signature:
    def __init__(self, key_id, algorithm, headers, signature):
        self.key_id = key_id
        self.algorithm = algorithm
        self.headers = headers
        self.signature = signature

        if self.algorithm not in ["rsa-sha256", "hs2019"]:
            logger.error(f"Unsupported algorithm {self.algorithm}")
            logger.error(self.signature)
            logger.error(self.headers)
            logger.error(self.key_id)

            raise Exception(f"Unsupported algorithm {self.algorithm}")

    def fields(self):
        return self.headers.split(" ")

    @staticmethod
    def from_signature_header(header):
        try:
            headers = header.split(",")
            headers = [x.split('="', 1) for x in headers]
            parsed = {x[0]: x[1].replace('"', "") for x in headers}

            return Signature(
                parsed["keyId"],
                parsed.get("algorithm", "rsa-sha256"),
                parsed["headers"],
                parsed["signature"],
            )
        except Exception:
            logger.error(f"failed to parse signature {header}")


def parse_signature_header(header):
    return Signature.from_signature_header(header)



import bleach

from bovine.activitystreams.utils.print import print_activity

from .collection_iterator import CollectionIterator


async def all_collection_elements(client, collection):
    result = []
    if isinstance(collection, str):
        collection = await client.proxy_element(collection)

    if "items" in collection:
        result = collection["items"]
        if isinstance(result, str):
            result = [result]

    if "first" in collection:
        return result + await all_collection_elements(client, collection["first"])

    if "next" in collection:
        return result + await all_collection_elements(client, collection["next"])

    return result


def short_version_of_object(obj):
    if "object" in obj:
        obj = obj["object"]

    for key in ["name", "summary", "content", "id"]:
        if key in obj and obj[key]:
            return bleach.clean(obj[key], tags=[], strip=True)

    return "--- unknown  ---"


class CollectionHelper:
    def __init__(self, collection_id, actor):
        self.collection_id = collection_id
        self.actor = actor

        self.basic_information = None
        self.items = None
        self.next_items = None
        self.item_index = None

        self.element_cache = {}

    async def refresh(self):
        if self.basic_information is None:
            self.basic_information = await self.actor.get(self.collection_id)

        if "first" not in self.basic_information:
            return

        response = await self.actor.get(self.basic_information["first"])

        self.items = response["orderedItems"]
        self.next_items = response["next"]
        self.item_index = 0

    async def get_element(self, element):
        if not isinstance(element, str):
            return element

        if element in self.element_cache:
            return self.element_cache[element]

        result = await self.actor.proxy_element(element)

        self.element_cache[element] = result

        return result

    async def next_item(self, do_print=False):
        if self.item_index >= len(self.items):
            response = await self.actor.get(self.next_items)
            self.items += response["orderedItems"]
            self.next_items = response["next"]

        result = self.items[self.item_index]
        self.item_index += 1

        result = await self.get_element(result)

        if do_print:
            print_activity(result)

        return result

    def iterate(self, max_number=10):
        return CollectionIterator(self, max_number)



import pytest

from bovine import BovineClient

from .collection_helper import all_collection_elements


@pytest.mark.skip("Network requests")
async def test_collections():
    remote = "https://indieweb.social/@tchambers/109993434359560479"
    async with BovineClient.from_file("h.toml") as client:
        result = await client.proxy_element(remote)

        replies = result["replies"]

        replies = await all_collection_elements(client, replies)

        assert isinstance(replies, list)



from datetime import datetime, timezone
from typing import Optional, Tuple


def webfinger_response_json(name: str, url: str, domain: str) -> dict:
    """helper to generate a webfinger response"""
    return {
        "subject": f"acct:{name}@{domain}",
        "links": [
            {
                "href": url,
                "rel": "self",
                "type": "application/activity+json",
            }
        ],
    }


def parse_fediverse_handle(account: str) -> Tuple[str, Optional[str]]:
    """Splits fediverse handle in name and domain"""
    if account[0] == "@":
        account = account[1:]

    if "@" in account:
        return tuple(account.split("@", 1))
    return account, None


def now_isoformat() -> str:
    return (
        datetime.now(tz=timezone.utc).replace(microsecond=0, tzinfo=None).isoformat()
        + "Z"
    )
